# Copyright (C) 2008  Nickolas Fotopoulos
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This module is intended to store generic, reusable, sub-classable plot classes
to minimize formulaic copying and pasting.
"""

__author__ = "Nickolas Fotopoulos <nvf@gravity.phys.uwm.edu>"

import itertools
from mpl_toolkits.mplot3d import Axes3D

import numpy
import pylab

from glue import iterutils

from pylal import viz

# general defaults
pylab.rc("lines", markersize=12)
pylab.rc("text", usetex=True)

# Utility function
def float_to_latex(x, format="%.2g"):
    """
    Convert a floating point number to a latex representation.  In particular,
    scientific notation is handled gracefully: e -> 10^
    """
    base_str = format % x
    if "e" not in base_str:
        return base_str
    mantissa, exponent = base_str.split("e")
    exponent = exponent.lstrip("0+")
    if mantissa == "1":
        return r"10^{%s}" % exponent
    else:
        return r"%s\times 10^{%s}" % (mantissa, exponent)

##############################################################################
# abstract classes

class BasicPlot(object):
    """
    A very default meta-class to almost any plot you might want to make.
    It provides basic initialization, a savefig method, and a close method.
    It is up to developers to subclass BasicPlot and fill in the add_content()
    and finalize() methods.
    """
    def __init__(self, xlabel="", ylabel="", title=""):
        """
        Basic plot initialization.  A subclass can override __init__ and call
        this one (plotutils.BasicPlot.__init__(self, *args, **kwargs)) and
        then initialize variables to hold data to plot and labels.
        """
        self.fig = pylab.figure()
        self.ax = self.fig.add_subplot(111)

        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(ylabel)
        self.ax.set_title(title)

        self.ax.grid(True)

    def add_content(self, data, label="_nolabel_"):
        """
        Stub.  Replace with a method that appends values or lists of values
        to self.data_sets and appends labels to self.data_labels.  Feel free
        to accept complicated inputs, but try to store only the raw numbers
        that will enter the plot.
        """
        raise NotImplementedError

    def finalize(self):
        """
        Stub.  Replace with a function that creates and makes your plot
        pretty.  Do not do I/O here.
        """
        raise NotImplementedError

    def savefig(self, *args, **kwargs):
        self.fig.savefig(*args, **kwargs)

    def close(self):
        """
        Close the plot and release its memory.
        """
        pylab.close(self.fig)
        
    def add_legend_if_labels_exist(self, *args, **kwargs):
        """
        Create a legend if there are any non-trivial labels.
        """
        for plot_kwargs in self.kwarg_sets:
            if "label" in plot_kwargs and \
                not plot_kwargs["label"].startswith("_"):
                self.ax.legend(*args, **kwargs)
                return

##############################################################################
# utility functions

def default_colors():
    """
    An infinite iterator of some default colors.
    """
    return itertools.cycle(('b', 'g', 'r', 'c', 'm', 'y', 'k'))

def default_symbols():
    """
    An infinite iterator of some default symbols.
    """
    return itertools.cycle(('x', '^', 'D', 'H', 'o', '1', '+'))

def determine_common_bin_limits(data_sets, default_min=0, default_max=0):
    """
    Given a some nested sequences (e.g. list of lists), determine the largest
    and smallest values over the data sets and determine a common binning.
    """
    max_stat = max(list(iterutils.flatten(data_sets)) + [-numpy.inf])
    min_stat = min(list(iterutils.flatten(data_sets)) + [numpy.inf])
    if numpy.isinf(-max_stat):
        max_stat = default_max
    if numpy.isinf(min_stat):
        min_stat = default_min
    return min_stat, max_stat

def method_callable_once(f):
    """
    Decorator to make a method complain if called more than once.
    """
    def _new(self, *args, **kwargs):
        attr = "_" + f.__name__ + "_already_called"
        if hasattr(self, attr) and getattr(self, attr):
            raise ValueError, f.__name__ + " can only be called once"
        setattr(self, attr, True)
        return f(self, *args, **kwargs)
    _new.__doc__ == f.__doc__
    _new.__name__ = f.__name__
    return _new

##############################################################################
# generic, but usable classes

class SimplePlot(BasicPlot):
    """
    Exactly what you get by calling pylab.plot(), but with the handy extras
    of the BasicPlot class.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.x_data_sets = []
        self.y_data_sets = []
        self.kwarg_sets = []

    def add_content(self, x_data, y_data, **kwargs):
        self.x_data_sets.append(x_data)
        self.y_data_sets.append(y_data)
        self.kwarg_sets.append(kwargs)

    @method_callable_once
    def finalize(self, loc=0):
        # make plot
        for x_vals, y_vals, plot_kwargs in \
            itertools.izip(self.x_data_sets, self.y_data_sets, self.kwarg_sets):
            self.ax.plot(x_vals, y_vals, **plot_kwargs)

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist(loc=loc)

        # decrement reference counts
        del self.x_data_sets
        del self.y_data_sets
        del self.kwarg_sets

class BarPlot(BasicPlot):
    """
    A simple vertical bar plot.  Bars are centered on the x values and have
    height equal to the y values.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.x_data_sets = []
        self.y_data_sets = []
        self.kwarg_sets = []

    def add_content(self, x_data, y_data, **kwargs):
        self.x_data_sets.append(x_data)
        self.y_data_sets.append(y_data)
        self.kwarg_sets.append(kwargs)

    @method_callable_once
    def finalize(self, orientation="vertical"):
        # make plot
        for x_vals, y_vals, plot_kwargs in \
            itertools.izip(self.x_data_sets, self.y_data_sets,
                           self.kwarg_sets):
            plot_kwargs.setdefault("align", "center")
            # FIXME: linewidth is not a valid kwarg in matplotlib 0.87.7
            # Reenable once clusters upgrade to CentOS 5.  Until then,
            # all bars have thick, black borders.
            #plot_kwargs.setdefault("linewidth", 0)
            plot_kwargs.setdefault("orientation", orientation)
            self.ax.bar(x_vals, y_vals, **plot_kwargs)

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist()

        # decrement reference counts
        del self.x_data_sets
        del self.y_data_sets
        del self.kwarg_sets

class VerticalBarHistogram(BasicPlot):
    """
    Histogram data sets with a common binning, then make a vertical bar plot.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.data_sets = []
        self.kwarg_sets = []

    def add_content(self, data, **kwargs):
        self.data_sets.append(data)
        self.kwarg_sets.append(kwargs)

    @method_callable_once
    def finalize(self, num_bins=20, normed=False):
        # determine binning
        min_stat, max_stat = determine_common_bin_limits(self.data_sets)
        bins = numpy.linspace(min_stat, max_stat, num_bins + 1, endpoint=True)

        # determine bar width; gets silly for more than a few data sets
        width = (1 - 0.1 * len(self.data_sets)) * (bins[1] - bins[0])

        # make plot
        legends = []
        plot_list = []
        for i, (data_set, plot_kwargs) in \
            enumerate(itertools.izip(self.data_sets, self.kwarg_sets)):
            # set default values
            plot_kwargs.setdefault("alpha", 0.6)
            plot_kwargs.setdefault("align", "center")
            plot_kwargs.setdefault("width", width)

            # make histogram
            y, x = numpy.histogram(data_set, bins=bins, normed=normed, new=True)
            x = x[:-1]

            # stagger bins for pure aesthetics
            x += 0.1 * i * max_stat / num_bins

            # plot
            plot_item = self.ax.bar(x, y, **plot_kwargs)

            # add legend and the right plot instance
            # for creating the correct labels!
            if "label" in plot_kwargs and \
                   not plot_kwargs["label"].startswith("_"):

                legends.append(plot_kwargs["label"])
                plot_list.append(plot_item[0])                

        # add legend if there are any non-trivial labels
        self.ax.legend(plot_list, legends)

        # decrement reference counts
        del self.data_sets
        del self.kwarg_sets

class NumberVsBinBarPlot(BasicPlot):
    """
    Make a bar plot in which the width and placement of the bars are set
    by the given bins.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.bin_sets = []
        self.value_sets = []
        self.kwarg_sets = []

    def add_content(self, bins, values, **kwargs):
        if len(bins) != len(values):
            raise ValueError, "length of bins and values do not match"

        self.bin_sets.append(bins)
        self.value_sets.append(values)
        self.kwarg_sets.append(kwargs)

    @method_callable_once
    def finalize(self, orientation="vertical"):
        for bins, values, plot_kwargs in itertools.izip(self.bin_sets,
            self.value_sets, self.kwarg_sets):
            x_vals = bins.centres()

            # prevent each bar from getting a separate legend entry
            label = "_nolegend_"
            if "label" in plot_kwargs:
                label = plot_kwargs["label"]
                del plot_kwargs["label"]

            # set default
            plot_kwargs.setdefault("align", "center")

            if orientation == "vertical":
                plot_kwargs.setdefault("width", bins.upper() - bins.lower())
                patches = self.ax.bar(x_vals, values, **plot_kwargs)
            elif orientation == "horizontal":
                plot_kwargs.setdefault("height", bins.upper() - bins.lower())
                patches = self.ax.barh(x_vals, values, **plot_kwargs)
            else:
                raise ValueError, orientation + " must be 'vertical' " \
                    "or 'horizontal'"

            # prevent each bar from getting a separate legend entry
            if len(patches) > 0:
                patches[0].set_label(label)

        pylab.axis('tight')

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist()

        # decrement reference counts
        del self.bin_sets
        del self.value_sets
        del self.kwarg_sets

class CumulativeHistogramPlot(BasicPlot):
    """
    Cumulative histogram of foreground that also has a shaded region,
    determined by the mean and standard deviation of the background
    population coincidence statistics.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.fg_data_sets = []
        self.fg_kwarg_sets = []
        self.bg_data_sets = []
        self.bg_kwargs = {}

    def add_content(self, fg_data_set, **kwargs):
        self.fg_data_sets.append(fg_data_set)
        self.fg_kwarg_sets.append(kwargs)

    def add_background(self, bg_data_sets, **kwargs):
        self.bg_data_sets.extend(bg_data_sets)
        self.bg_kwargs = kwargs

    @method_callable_once
    def finalize(self, num_bins=20, normalization=1):
        epsilon = 1e-8

        # determine binning
        min_stat, max_stat = determine_common_bin_limits(\
            self.fg_data_sets + self.bg_data_sets)
        bins = numpy.linspace(min_stat, max_stat, num_bins + 1, endpoint=True)
        dx = bins[1] - bins[0]

        # plot foreground
        for data_set, plot_kwargs in \
            itertools.izip(self.fg_data_sets, self.fg_kwarg_sets):
            # make histogram
            y, x = numpy.histogram(data_set, bins=bins, new=True)
            y = y[::-1].cumsum()[::-1]
            x = x[:-1]

            # plot
            y = numpy.array(y, dtype=numpy.float32)
            y[y <= epsilon] = epsilon
            self.ax.plot(x + dx/2, y*normalization, **plot_kwargs)

        # shade background region
        if len(self.bg_data_sets) > 0:
            # histogram each background instance separately and take stats
            hist_sum = numpy.zeros(len(bins), dtype=float)
            sq_hist_sum = numpy.zeros(len(bins), dtype=float)
            for instance in self.bg_data_sets:
                # make histogram
                y, x = numpy.histogram(instance, bins=bins, new=False)
                y = y[::-1].cumsum()[::-1]
                hist_sum += y
                sq_hist_sum += y*y

            # get statistics
            N = len(self.bg_data_sets)
            means = hist_sum / N
            stds = numpy.sqrt((sq_hist_sum - hist_sum*means) / (N - 1))

            # plot mean
            means[means <= epsilon] = epsilon
            self.ax.plot(x + dx/2, means*normalization, 'r+', **self.bg_kwargs)

            # shade in the area
            if "label" in self.bg_kwargs:
                self.bg_kwargs["label"] = r"$\mu_\mathrm{%s}$" \
                    % self.bg_kwargs["label"]
            self.bg_kwargs.setdefault("alpha", 0.3)
            self.bg_kwargs.setdefault("facecolor", "y")
            upper = means + stds
            lower = means - stds
            lower[lower <= epsilon] = epsilon
            tmp_x, tmp_y = viz.makesteps(bins, upper, lower)
            self.ax.fill(tmp_x, tmp_y*normalization, **self.bg_kwargs)

        # make semilogy plot
        self.ax.set_yscale("log")

        # adjust plot range
        self.ax.set_xlim((0.9 * min_stat, 1.1 * max_stat))
        possible_ymins = [0.6]
        if len(self.bg_data_sets) > 0:
            possible_ymins.append(0.6 / N)
        else:
            possible_ymins.append(0.6 * normalization)
        self.ax.set_ylim(min(possible_ymins))

        # add legend if there are any non-trivial labels
        self.kwarg_sets = self.fg_kwarg_sets
        self.add_legend_if_labels_exist()

        # decrement reference counts
        del self.kwarg_sets
        del self.fg_data_sets
        del self.fg_kwarg_sets
        del self.bg_data_sets
        del self.bg_kwargs


class ImagePlot(BasicPlot):
    """
    The equivalent of pylab.imshow(), but with the BasicPlot niceties and
    some defaults that are more in tune with what a scientist wants --
    origin="lower", requiring x and y bins so that we can label axes
    correctly, and a colorbar.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.image = None

    def add_content(self, image, x_bins, y_bins):
        """
        Add a given image to this plot.

        @param image: two-dimensional numpy array to plot
        @param x_bins: pylal.rate.Bins instance describing the x binning
        @param y_bins: pylal.rate.Bins instance describing the y binning
        """
        if self.image is not None:
            raise ValueError, "can only plot one image"
        if image.ndim != 2:
            raise ValueError, "require 2-D array"
        self.image = image
        self.x_bins = x_bins
        self.y_bins = y_bins

    @method_callable_once
    def finalize(self, colorbar=True, **kwargs):
        if self.image is None:
            raise ValueError, "nothing to finalize"

        extent = [self.x_bins.lower()[0], self.x_bins.upper()[-1],
                  self.y_bins.lower()[0], self.y_bins.upper()[-1]]

        kwargs.setdefault("origin", "lower")
        kwargs.setdefault("interpolation", "nearest")

        im = self.ax.imshow(self.image, extent=extent, **kwargs)

        pylab.axis('tight')

        if colorbar:
            self.fig.colorbar(im)

        # XXX: hack our ticks into log space; imshow doesn't support log scales
        def log_transform(lin_range):
            """
            Return the logarithmic ticks and labels corresponding to the
            input lin_range.
            """
            log_range = numpy.log10(lin_range)
            slope = (lin_range[1] - lin_range[0]) \
                  / (log_range[1] - log_range[0])
            inter = lin_range[0] - slope * log_range[0]
            tick_range = [tick for tick in range(int(log_range[0] - 1.0),
                                                 int(log_range[1] + 1.0))\
                          if tick >= log_range[0] and tick<=log_range[1]]
            ticks = [inter + slope * tick for tick in tick_range]
            labels = ["${10^{%d}}$" % tick for tick in tick_range]
            return ticks, labels
        from pylal import rate
        if isinstance(self.x_bins, rate.LogarithmicBins):
            xticks, xlabels = log_transform(self.ax.get_xlim())
            self.ax.set_xticks(xticks)
            self.ax.set_xticklabels(xlabels)
        if isinstance(self.y_bins, rate.LogarithmicBins):
            yticks, ylabels = log_transform(self.ax.get_ylim())
            self.ax.set_yticks(yticks)
            self.ax.set_yticklabels(ylabels)


class FillPlot(BasicPlot):
    """
    Given a list of vertices (passed by x-coords and y-coords), fill the
    regions (by default with successively darker gray shades).
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.x_coord_sets = []
        self.y_coord_sets = []
        self.kwarg_sets = []
        self.shades = []

    def add_content(self, x_coords, y_coords, shade=None, **kwargs):
        if len(x_coords) != len(y_coords):
            raise ValueError, "x and y coords have different length"
        if iterutils.any(s is None for s in self.shades) and shade is not None \
        or iterutils.any(s is not None for s in self.shades) and shade is None:
            raise ValueError, "cannot mix explicit and automatic shading"

        self.x_coord_sets.append(x_coords)
        self.y_coord_sets.append(y_coords)
        self.kwarg_sets.append(kwargs)
        self.shades.append(shade)

    @method_callable_once
    def finalize(self):
        # fill in default shades if necessary
        if iterutils.any(s is None for s in self.shades):
            n = len(self.shades)
            grays = numpy.linspace(0, 1, n, endpoint=False)[::-1]
            self.shades = numpy.vstack((grays, grays, grays)).T

        # plot
        for x, y, s, plot_kwargs in zip(self.x_coord_sets, self.y_coord_sets,
                              self.shades, self.kwarg_sets):
            self.ax.fill(x, y, facecolor=s, **plot_kwargs)

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist(loc=0)

class SixStripSeriesPlot(BasicPlot):
    """
    Given a time- or frequency-series, plot it across six horizontal axes,
    stacked on top of one another.  This is good for representing a
    high-resolution one-dimensional data set.  To support missing data,
    we require x and y coordinates for each data set.
    """
    def __init__(self, xlabel="", ylabel="", title=""):
        self.fig = pylab.figure(figsize=(5.54, 7.5))
        self.title = title

        # do not want to create axes yet
        self.xlabel = xlabel
        self.ylabel = ylabel

        self.x_coord_sets = []
        self.y_coord_sets = []
        self.kwarg_sets = []

    def add_content(self, x_coords, y_coords, **kwargs):
        if len(x_coords) != len(y_coords):
            raise ValueError, "x and y coords have different length"
        if len(self.kwarg_sets) and \
            (iterutils.any("color" in kw for kw in self.kwarg_sets)
             ^ ("color" in kwargs)):
            raise ValueError, "cannot mix explicit and automatic coloring"

        self.x_coord_sets.append(x_coords)
        self.y_coord_sets.append(y_coords)
        self.kwarg_sets.append(kwargs)

    @method_callable_once
    def finalize(self, yscale="linear"):
        for kw, c in zip(self.kwarg_sets, default_colors()):
            kw.setdefault("color", c)

        min_x, max_x = determine_common_bin_limits(self.x_coord_sets)

        numaxes = 6  # This is hardcoded below.  Change at your own risk.
        ticks_per_axis = 6 # Hardcoded, but you can change it if it looks bad.
        fperaxis = (max_x - min_x) / numaxes
        freq_boundaries = numpy.linspace(min_x, max_x, numaxes + 1)
        freq_ranges = zip(freq_boundaries[:-1], freq_boundaries[1:])

        # attempt to put ticks at every multiple of 10 unless there are too few
        # or too many
        tickspacing = int(numpy.ceil(fperaxis / ticks_per_axis / 10)) * 10
        if abs(fperaxis / tickspacing - ticks_per_axis) > 2:
            tickspacing = int(numpy.ceil(fperaxis / ticks_per_axis))

        # iterate over axes
        for j, freq_range in enumerate(freq_ranges):
            # create one of the 6 axes
            ax = self.fig.add_axes([.12, .84-.155*j, .86, 0.124])

            for x_coords, y_coords, kwarg_set in zip(self.x_coord_sets,
                self.y_coord_sets, self.kwarg_sets):
                # just look at the region relevant to our axis
                ind = (x_coords >= freq_range[0]) & (x_coords < freq_range[1])
                x = x_coords[ind]
                y = y_coords[ind]

                # add data to axes
                ax.plot(x, y, **kwarg_set)

                # fix the limits and ticks
                ax.set_xlim(freq_range)

                mintick = int(numpy.ceil(freq_range[0] / tickspacing)) \
                    * tickspacing
                maxtick = int(numpy.floor(freq_range[1] / tickspacing)) \
                    * tickspacing
                ax.set_xticks(xrange(mintick, maxtick+1, tickspacing))

                ax.set_ylabel(self.ylabel)
                ax.set_yscale(yscale)
                ax.grid(True)

        # label bottom row
        ax.set_xlabel(self.xlabel)

        # title top row
        self.fig.axes[0].set_title(self.title)

        # apply common y limits
        y_lims = [ax.get_ylim() for ax in self.fig.axes]
        new_y_lims = determine_common_bin_limits(y_lims)
        for ax in self.fig.axes:
            ax.set_ylim(new_y_lims)

class ROCPlot(BasicPlot):
    """
    Plot the receiver operating characteristic (ROC) based on the foreground
    and background values from given techniques.  For example, to compare
    SNR vs IFAR, do something like:

    plot = ROCPlot("FAP", "EFF", "ROC IFAR vs SNR")
    plot.add_content(ifar_bg, ifar_fg, label="IFAR")
    plot.add_content(snr_bg, snr_fg, label="SNR")
    plot.finalize()
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.bg_sets = []
        self.fg_sets = []
        self.eff_weight_sets = []
        self.kwarg_sets = []

    def add_content(self, bg, fg, eff_weight=None, **kwargs):
        """
        Enter a particular technique's background, foreground, and efficiency
        weights.  These should be one-dimensional arrays with the values of
        of your statistics for backgrounds and foregrounds.  Eff_weight
        are weights on the efficiency, useful if, say, you have a different
        prior than your injection set reflects.
        """
        if eff_weight is not None and len(fg) != len(eff_weight):
            raise ValueError, "efficiency weights and foreground values "\
                "must be the same length"
        self.bg_sets.append(bg)
        self.fg_sets.append(fg)
        self.eff_weight_sets.append(eff_weight)
        self.kwarg_sets.append(kwargs)

    @method_callable_once
    def finalize(self, loc=0):
        for bg, fg, weights, kwargs in itertools.izip(self.bg_sets,
            self.fg_sets, self.eff_weight_sets, self.kwarg_sets):
            # sort, keeping the weights and fg values together
            bg = numpy.array(bg)  # always copies
            bg.sort()
            fg = numpy.array(fg)  # always copies
            if weights is not None:
                weights = numpy.array(weights)[fg.argsort()]
            fg.sort()

            # calculate false alarm probability and efficiency
            FAP = 1 - numpy.arange(len(bg), dtype=float) / len(bg)
            if weights is not None:
                EFF = weights[::-1].cumsum()[::-1]
            else:
                EFF = 1 - numpy.arange(len(fg), dtype=float) / len(fg)

            # now find the efficiency *at* each false alarm probability
            # if louder than loudest event, then efficiency is 0
            EFF_ind = numpy.array([fg.searchsorted(x) for x in bg])
            fg_too_loud = EFF_ind == len(fg)
            EFF_ind[fg_too_loud] = 0  # dummy index
            EFF_by_FAP = EFF[EFF_ind]
            EFF_by_FAP[fg_too_loud] = 0.

            # plot!
            self.ax.plot(FAP, EFF_by_FAP, **kwargs)

        # make it pretty
        self.ax.grid(True)
        self.ax.set_xlim((0, 1))
        self.ax.set_ylim((0, 1))

        # resize figure to make axes square
        fig_side = min(self.fig.get_size_inches())
        self.fig.set_size_inches(fig_side, fig_side)

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist(loc=loc)

        # decrement reference counts
        del self.fg_sets
        del self.bg_sets
        del self.eff_weight_sets
        del self.kwarg_sets

class QQPlot(BasicPlot):
    """
    Plot the global rank versus the self rank, like a Q-Q plot, i.e.
    differences between the probability distribution of a
    statistical population to a test population.
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.fg_data = []
        self.bg_data = []
        self.fg_kwargs = []
        self.bg_kwargs = []
        self.kwargs = None

    def add_content(self, *args):
        raise NotImplementedError, "A function 'add_content' is not "\
              "implemented for QQPlot. "\
              "Please use the functions 'add_fg' and 'add_bg'."

    def _compute_quantiles(self, fg, bg):
        """
        Calculates the sorted quantiles to be plotted.
        """
        from scipy import stats

        N_FG = len(fg)
        N_BG = len(bg)
        N_tot = N_BG + N_FG

        # global rank vs self rank; very much like a Q-Q plot
        bg_self_quantile = stats.rankdata(bg).astype(float) / N_BG
        fg_self_quantile = stats.rankdata(fg).astype(float) / N_FG

        popAB = numpy.concatenate((fg, bg))
        popAB_ranked = stats.rankdata(popAB).astype(float)
        fg_global_quantile = popAB_ranked[:N_FG] / N_tot
        bg_global_quantile = popAB_ranked[N_FG:] / N_tot

        fg_self_quantile.sort()
        bg_self_quantile.sort()
        fg_global_quantile.sort()
        bg_global_quantile.sort()

        return fg_self_quantile, bg_self_quantile, fg_global_quantile, bg_global_quantile

    def add_fg(self, data, **kwargs):
        """
        Add the foreground data and the kwargs to a list for the
        """
        self.fg_data.append(data)
        self.fg_kwargs.append(kwargs)

    def add_bg(self, data, **kwargs):
        """
        Add the foreground data and the kwargs to a list for the
        """
        self.bg_data.append(data)
        self.bg_kwargs.append(kwargs)

    @method_callable_once
    def finalize(self, loc=0):

        if len(self.fg_data)!=len(self.bg_data):
            raise ValueError, "You have to add the same number of foreground data"\
                  " as background data. But you have specified %d sets of foreground data but %d"\
                  " sets of background data. " % ( len(self.fg_data), len(self.bg_data))

        for fg, bg, fg_kwargs, bg_kwargs in itertools.izip(self.fg_data, self.bg_data, \
                                                           self.fg_kwargs, self.bg_kwargs):

            fg_self, bg_self, fg_global, bg_global = \
                     self._compute_quantiles(fg, bg)

            self.ax.plot(bg_self, bg_global, **bg_kwargs)
            self.ax.plot(fg_self, fg_global, **fg_kwargs)

        # make it pretty
        self.ax.grid(True)

        # add legend if there are any non-trivial labels
        self.ax.legend(loc=4)

        # decrement reference counts
        del self.fg_data
        del self.bg_data
        del self.fg_kwargs
        del self.bg_kwargs
        del self.kwargs

class SimpleMapPlot(BasicPlot):
    """
    Class to create a clickable map html page. 
    """
    def __init__(self, *args, **kwargs):
        BasicPlot.__init__(self, *args, **kwargs)
        self.x_data_sets = []
        self.y_data_sets = []
        self.link_list = []        
        self.kwarg_sets = []

        self.click_size = 5
        
        self.click_x = []
        self.click_y = []
        self.click_link = []
        
    def set_click_size(self, size):
        """
        Sets the size of the area around a point that can be clicked at.
        """
        self.click_size = size

    def add_content(self, x_data, y_data, link, **kwargs):
        self.x_data_sets.append(x_data)
        self.y_data_sets.append(y_data)
        self.link_list.append(link)
        self.kwarg_sets.append(kwargs)

        
    @method_callable_once
    def finalize(self, loc=0):
        
        # make plot
        for x_vals, y_vals, plot_kwargs in \
            itertools.izip(self.x_data_sets, self.y_data_sets, self.kwarg_sets):
            self.ax.plot(x_vals, y_vals, **plot_kwargs)

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist(loc=loc)

          
    def rescale(self, dpi):
        """
        Calculate the rescaling to map the point coordinates
        to the actual coordinates on the image.
        """

        # set the dpi used for the savefig
        if dpi is None:
            dpi = pylab.rcParams['savefig.dpi']
        self.ax.get_figure().set_dpi(dpi)

        # get the full extend of the final image
        _, _, width, height = self.fig.bbox.extents

        # set the pixels for each point that can be clicked
        self.click_x = []
        self.click_y = []
        self.click_link = []
        for xvec, yvec, link in \
                itertools.izip(self.x_data_sets, self.y_data_sets, self.link_list):
            # skip if no link is associated with these point(s)
            if link is None:
                continue
            
            for point in zip(xvec, yvec):

                # transform the data coordinates into piel coordinates
                pixel = self.ax.transData.transform(point)

                # save the new coordinates. The y-coordinate is just
                # the other way around. 
                self.click_x.append(pixel[0])
                self.click_y.append(height - pixel[1])
                self.click_link.append(link)


    def create_html(self, plotname, dpi = None):
        """
        Create the html file with the name of the plot
        as the only additional input information needed.
        If the actual plot is saved with a different
        dpi than the standard one, it must be specified here!!
        """
                                  
        # points need to be rescaled first
        self.rescale(dpi)
        
        # get the full extend of the final image
        _, _, width, height = self.fig.bbox.extents
        
        # create the map-page
        page = ''
        page += '<IMG src="%s" width=%dpx '\
            'usemap="#map">' % ( plotname, width) 
        page +=  '<MAP name="map"> <P>' 
        n=0
        for px, py, link in zip( self.click_x, self.click_y, self.click_link):
            n+=1
            page +=  '<area href="%s" shape="circle" '\
                'coords="%d, %d, 5"> Point%d</a>' %\
                ( link, px, py, n) 
        page += '</P></MAP></OBJECT><br>'
        page += "<hr/>"

        # return the html code, to be saved to a file or embedded in a
        # larger html page
        return page



plot3D_head = """
  <style type="text/css">
    .rotationViewer {
      position:relative;
      width:640px;
      height:378px;
      border-style:solid;
      border-width:1px;
      margin:auto;
      margin-bottom:10px;
      cursor:pointer;
    }
    .rotationViewer img {
      position:absolute;
      visibility:hidden;
      left:0;
      top:0;
      width:100%;
      height:100%;
    }
  </style>
"""

plot3D_body = """
<script type="text/javascript" src="PATHTOUIZE/Uize.js"></script>

<div class="main">
  <div id="page_rotationViewer" class="rotationViewer insetBorderColor"></div>
</div>

<!-- JavaScript code to make the static bar HTML "come alive" -->
<!-- Based on http://www.uize.com/examples/3d-rotation-viewer.html -->
<script type="text/javascript">

Uize.module ({
  required:[
    'UizeDotCom.Page.Example.library',
    'UizeDotCom.Page.Example',
    'Uize.Widget.Drag',
    'Uize.Fade.xFactory',
    'Uize.Curve.Rubber',
    'Uize.Curve.Mod'
  ],
  builder:function () {
    /*** create the example page widget ***/
      var page = window.page = Uize.Widget.Page  ({evaluator:function (code) {eval (code)}}); 

    /*** configuration variables ***/
      var
        totalFrames = TOTALFRAMES,
        frameUrlTemplate = 'URLTEMPLATE'
      ;

    /*** state variables ***/
      var
        rotation = 0,
        lastFrameNo = -1,
        dragStartRotation
      ;

    /*** create the Uize.Widget.Drag instance ***/
      var rotationViewer = page.addChild (
        'rotationViewer',
        Uize.Widget.Drag,
        {
          cancelFade:{duration:5000,curve:Uize.Curve.Rubber.easeOutBounce ()},
          releaseTravel:function (speed) {
            var
              deceleration = 5000, // measured in pixels/s/s
              duration = speed / deceleration
            ;
            return {
              duration:duration,
              distance:Math.round (speed * duration / 2),
              curve:function (_value) {return 1 - (_value = 1 - _value) * _value}
            };
          },
          html:function (input) {
            var
              htmlChunks = [],
              frameNodeIdPrefix = input.idPrefix + '-frame'
            ;
            for (var frameNo = 0; ++frameNo <= totalFrames;) {
              htmlChunks.push (
                '<img' +
                  ' id="' + frameNodeIdPrefix + frameNo + '"' +
                  ' src="' + Uize.substituteInto (frameUrlTemplate,{frame:(frameNo < 10 ? '0' : '') + frameNo}) +'"' +
                '/>'
              );
            }
            return htmlChunks.join ('');
          },
          built:false
        }
      );

    /*** wire up the drag widget with events for updating rotation degree ***/
      function updateRotation (newRotation) {
        rotation = ((newRotation % 360) + 360) % 360;
        var frameNo = 1 + Math.round (rotation / 360 * (totalFrames - 1));
        if (frameNo != lastFrameNo) {
          rotationViewer.showNode ('frame'+ lastFrameNo,false);
          rotationViewer.showNode ('frame'+ (lastFrameNo = frameNo));
        }
      }
      rotationViewer.wire ({
        'Drag Start':function () {dragStartRotation = rotation},
        'Drag Update':function (e) {updateRotation (dragStartRotation - e.source.eventDeltaPos [0] / 2.5)}
      });

    /*** function for animating spin ***/
      function spin (degrees,duration,curve) {
        Uize.Fade.fade (updateRotation,rotation,rotation + degrees,duration,{quantization:1,curve:curve});
      }

    /*** initialization ***/
      Uize.Node.wire (window,'load',function () {spin (360,2700,Uize.Curve.easeInOutPow (4))});

    /*** wire up the page widget ***/
      page.wireUi ();
  }
});

</script>
"""


class Plot3D(BasicPlot):
    """
    Exactly what you get by calling pylab.plot(), but with the handy extras
    of the BasicPlot class.
    """
    def __init__(self, *args, **kwargs):

      
        BasicPlot.__init__(self)
        self.x_data_sets = []
        self.y_data_sets = []
        self.z_data_sets = []        
        self.kwarg_sets = []

        # need the 3D axes
        self.ax =  Axes3D(self.fig)

        # set the labels on the new 3D axis
        self.ax.set_xlabel(args[0])
        self.ax.set_ylabel(args[1])
        self.ax.set_zlabel(args[2])
        self.ax.set_title(args[3])

        self.ax.grid(True)
        
    def add_content(self, x_data, y_data, z_data, **kwargs):
        self.x_data_sets.append(x_data)
        self.y_data_sets.append(y_data)
        self.z_data_sets.append(z_data)        
        self.kwarg_sets.append(kwargs)


    @method_callable_once
    def finalize(self, plot_basename, n=50, dpi=50, loc=0, plot_dir = '.',\
                     js_dir = 'https://ldas-jobs.phys.uwm.edu/~dietz/js'):

        # check input argument
        if n>100:
            raise ValueError("The 3D code cannot handle more than 100 frames currently.")

        fig = pylab.figure()
        ax = Axes3D(fig)
        
        # create plot
        for x_vals, y_vals, z_vals, plot_kwargs in \
            itertools.izip(self.x_data_sets, self.y_data_sets, \
                           self.z_data_sets, self.kwarg_sets):
            self.ax.scatter(x_vals, y_vals, z_vals, **plot_kwargs)

        # add legend if there are any non-trivial labels
        self.add_legend_if_labels_exist(loc=loc)

        # loop over a full circle with n steps
        dphi = 360.0/float(n)
        for i, angle in enumerate(numpy.arange(0.0, 360.0, dphi)):
            
            # make the rotation of the image
            self.ax.view_init(30.0, angle)

            # create the rotated image in the ploting directory
            picname = '%s/%s-%02d'%(plot_dir, plot_basename, i)
            self.savefig(picname+'.png',dpi=dpi)

            
        # decrement reference counts
        del self.x_data_sets
        del self.y_data_sets
        del self.z_data_sets        
        del self.kwarg_sets


        # produce the html output code
        plot_template = '%s/%s-[#frame].png'%(plot_dir, plot_basename)
        body = plot3D_body.replace('TOTALFRAMES',str(n))
        body = body.replace('URLTEMPLATE', plot_template)
        body = body.replace('PATHTOUIZE', js_dir)

        # return the html snippets
        return plot3D_head, body

    
###################################################
## unittest section
###################################################

import unittest

# --------------------------------------------
class TestSimpleMapPlot(unittest.TestCase):


    def test_plot(self):
        # set a different dpi for testing purposes
        pylab.rcParams.update({"savefig.dpi": 120})

        # define the SimpleMapPlot
        plot = SimpleMapPlot(r"Range [km]", r"Take off weight [t]", "4 engine planes")

        # define some data.
        # Note: The third item consists of two points (with the same link)
        # while the last data point has no link at all (link=None)
        range = [ [14600], [14800], [13000, 14800], [2800], [4800], [14000]]
        weight = [[365], [560], [374, 442], [46], [392], [50]]
        links = ['http://de.wikipedia.org/wiki/Airbus_A340',\
                'http://de.wikipedia.org/wiki/Airbus_A380',\
                'http://de.wikipedia.org/wiki/Boeing_747',\
                'http://de.wikipedia.org/wiki/Avro_RJ85',\
                'http://de.wikipedia.org/wiki/Antonow_An-124',\
                 None]
        markers = ['ro','bD','yx','cs','g^','k>']

        # set the data to plotutils
        for x,y,link, mark in zip(range, weight, links, markers):
            plot.add_content(x, y, link, color=mark[0], marker=mark[1])

        # finalize the plot
        plot.finalize()

        # save the plot
        plotname = 'plotutils_TestSimpleMapPlot.png'
        plot.savefig(plotname)

        # and here create the html click map and save the html file
        html = plot.create_html(plotname)
        f = file('plotutils_TestSimpleMapPlot.html','w')
        f.write(html)
        f.close()

# --------------------------------------------    
if __name__ == '__main__':

    
    unittest.main()
